@@ -1,5 +1,5 @@
 /******************************************************************************
- * Copyright 1994-2009,2010 by Thomas E. Dickey                               *
+ * Copyright 1994-2012,2013 by Thomas E. Dickey                               *
  * All Rights Reserved.                                                       *
  *                                                                            *
  * Permission to use, copy, modify, and distribute this software and its      *
@@ -20,7 +20,7 @@
  ******************************************************************************/
 
 #ifndef	NO_IDENT
-static const char *Id = "$Id: diffstat.c,v 1.54 2010/10/10 21:38:34 tom Exp $";
+static const char *Id = "$Id: diffstat.c,v 1.57 2013/04/16 00:29:29 tom Exp $";
 #endif
 
 /*
@@ -28,9 +28,24 @@ static const char *Id = "$Id: diffstat.c,v 1.54 2010/10/10 21:38:34 tom Exp $";
  * Author:	T.E.Dickey
  * Created:	02 Feb 1992
  * Modified:
+ *		15 Apr 2013, modify to accommodate output of "diff -q", which
+ *			     tells only if the files are different.  Work
+ *			     around the equivalent ambiguous message introduced
+ *			     in diffutils 2.8.4 and finally removed for 3.0
+ *		11 Feb 2013, add -K option.  Use strtol() to provide error
+ *			     checking of optarg values.
+ *		10 Feb 2013, document -b, -C, -s option in usage (patch by
+ *			     Tim Waugh, Red Hat #852770).  Improve pathname
+ *			     merging.
+ *		02 Jun 2012, fix for svn diff with spaces in path (patch by
+ *			     Stuart Prescott, Debian #675465).
+ *		03 Jan 2012, Correct case for "xz" suffix in is_compressed()
+ *			     (patch from Frederic Culot in FreeBSD ports).  Add
+ *			     "-R" option.  Improve dequoting of filenames in
+ *			     headers.
  *		10 Oct 2010, correct display of new files when -S/-D options
  *			     are used.  Remove the temporary directory on
- *			     error, introduced in 1.48+ (patch by Solaris
+ *			     error, introduced in 1.48+ (patch by Solar
  *			     Designer).
  *		19 Jul 2010, add missing "break" statement which left "-c"
  *			     option falling-through into "-C".
@@ -311,7 +326,7 @@ extern int optind;
 #define FMT_VERBOSE  4
 
 typedef enum comment {
-    Normal, Only, Binary
+    Normal, Only, OnlyLeft, OnlyRight, Binary, Differs, Either
 } Comment;
 
 #define MARKS 4			/* each of +, - and ! */
@@ -361,27 +376,29 @@ static const int colors[MARKS + 1] =
 static DATA *all_data;
 static const char *comment_opt = "";
 static char *path_opt = 0;
+static int count_files;		/* true if we count added/deleted files */
 static int format_opt = FMT_NORMAL;
+static int max_name_wide;	/* maximum amount reserved for filenames */
 static int max_width;		/* the specified width-limit */
 static int merge_names = 1;	/* true if we merge similar filenames */
 static int merge_opt = 0;	/* true if we merge ins/del as modified */
 static int min_name_wide;	/* minimum amount reserved for filenames */
-static int max_name_wide;	/* maximum amount reserved for filenames */
 static int names_only;		/* true if we list filenames only */
 static int num_marks = 3;	/* 3 or 4, according to "-P" option */
-static int show_colors;		/* true if showing SGR colors */
-static int show_progress;	/* if not writing to tty, show progress */
-static int summary_only = 0;	/* true if only summary line is shown */
 static int path_dest;		/* true if path_opt is destination (patched) */
 static int plot_width;		/* the amount left over for histogram */
 static int prefix_opt = -1;	/* if positive, controls stripping of PATHSEP */
+static int quiet = 0;		/* -q option */
+static int reverse_opt;		/* true if results are reversed */
 static int round_opt = 0;	/* if nonzero, round data for histogram */
+static int show_colors;		/* true if showing SGR colors */
+static int show_progress;	/* if not writing to tty, show progress */
+static int sort_names = 1;	/* true if we sort filenames */
+static int summary_only = 0;	/* true if only summary line is shown */
+static int suppress_binary = 0;	/* -b option */
 static int table_opt = 0;	/* if nonzero, write table rather than plot */
 static int trace_opt = 0;	/* if nonzero, write debugging information */
-static int sort_names = 1;	/* true if we sort filenames */
 static int verbose = 0;		/* -v option */
-static int quiet = 0;		/* -q option */
-static int suppress_binary = 0;	/* -b option */
 static long plot_scale;		/* the effective scale (1:maximum) */
 
 #ifdef HAVE_TSEARCH
@@ -393,6 +410,10 @@ static int prefix_len = -1;
 
 /******************************************************************************/
 
+#ifdef GCC_NORETURN
+static void failed(const char *) GCC_NORETURN;
+#endif
+
 static void
 failed(const char *s)
 {
@@ -570,9 +591,13 @@ delink(DATA * data)
     return 0;
 }
 
-/* like strncmp, but without the 3rd argument */
-static int
-match(const char *s, const char *p)
+/*
+ * Compare string 's' against a constant, returning either a pointer just
+ * past the matched part of 's' if it matches exactly, or null if a mismatch
+ * was found.
+ */
+static char *
+match(char *s, const char *p)
 {
     int ok = 0;
 
@@ -588,7 +613,7 @@ match(const char *s, const char *p)
 	    break;
 	}
     }
-    return ok;
+    return ok ? s : 0;
 }
 
 static int
@@ -646,7 +671,7 @@ decode_default(char *s,
 			*second_size = strtol(s, &next, 10) + 1 - *second;
 		    }
 		}
-		if (next != 0 && next != s && *next == '\0')
+		if (next != 0 && next != s && *next == EOS)
 		    rc = 1;
 		break;
 	    }
@@ -713,98 +738,200 @@ is_leaf(const char *theLeaf, const char *path)
     return 0;
 }
 
+static char *
+trim_datapath(DATA ** datap, size_t length, int *localp)
+{
+    char *target = (*datap)->name;
+
+#ifdef HAVE_TSEARCH
+    /*
+     * If we are using tsearch(), make a local copy of the data
+     * so we can trim it without interfering with tsearch's
+     * notion of the ordering of data.  That will create some
+     * spurious empty data, so we add the changed() macro in a
+     * few places to skip over those.
+     */
+    if (use_tsearch) {
+	char *trim = new_string(target);
+	trim[length] = EOS;
+	*datap = add_tsearch_data(trim, (*datap)->base);
+	target = (*datap)->name;
+	free(trim);
+	*localp = 1;
+    } else
+#endif
+	target[length] = EOS;
+
+    return target;
+}
+
+static size_t
+compare_tails(const char *target, const char *source, int *diff)
+{
+    size_t len1 = strlen(target);
+    size_t len2 = strlen(source);
+    size_t n;
+    size_t matched = 0;
+
+    *diff = 0;
+    for (n = 1; n <= len1 && n <= len2; n++) {
+	if (target[len1 - n] != source[len2 - n]) {
+	    *diff = (int) n;
+	    break;
+	}
+	if (source[len2 - n] == PATHSEP) {
+	    matched = n;
+	}
+    }
+    return matched;
+}
+
+/*
+ * The 'data' parameter points to the first of two markers, while
+ * 'path' is the pathname from the second marker.
+ *
+ * On the first call for
+ * a given file, the 'data' parameter stores no differences.
+ */
 static char *
 do_merging(DATA * data, char *path, int *freed)
 {
-    TRACE(("** do_merging(%s,%s) diffs:%d\n", data->name, path, HadDiffs(data)));
+    char *target = reverse_opt ? path : data->name;
+    char *source = reverse_opt ? data->name : path;
+    char *result = source;
+    int diff;
+
+    TRACE(("** do_merging(\"%s\",\"%s\") diffs:%d\n",
+	   data->name, path, HadDiffs(data)));
 
     *freed = 0;
-    if (!HadDiffs(data)) {	/* the data was the first of 2 markers */
-	if (is_leaf(data->name, path)) {
-	    TRACE(("** is_leaf: %s vs %s\n", data->name, path));
-	    *freed = delink(data);
-	} else if (can_be_merged(data->name)
-		   && can_be_merged(path)) {
-	    size_t len1 = strlen(data->name);
-	    size_t len2 = strlen(path);
-	    unsigned n;
+    if (!HadDiffs(data)) {
+
+	if (is_leaf(target, source)) {
+	    TRACE(("** is_leaf: \"%s\" vs \"%s\"\n", target, source));
+	    if (reverse_opt) {
+		TRACE((".. no action @%d\n", __LINE__));
+	    } else {
+		*freed = delink(data);
+	    }
+	} else if (can_be_merged(target)
+		   && can_be_merged(source)) {
+	    size_t len1 = strlen(target);
+	    size_t len2 = strlen(source);
 	    int matched = 0;
-	    int diff = 0;
 	    int local = 0;
 
 	    /*
-	     * Strip suffixes such as ".orig", ".bak", "~", etc.  The current
-	     * data for the merge is in 'path'.  The previous data (in
-	     * 'data->name') may also be a temporary filename (which would not
-	     * be merged since it has no apparent relationship to the current).
+	     * If the source/target differ only by some suffix, e.g., ".orig"
+	     * or ".bak", strip that off.  The target may may also be a
+	     * temporary filename (which would not be merged since it has no
+	     * apparent relationship to the current).
 	     */
 	    if (len1 > len2) {
-		if (!strncmp(data->name, path, len2)) {
-		    TRACE(("** trimming data '%s' to '%.*s'\n",
-			   data->name, (int) len2, data->name));
-		    len1 = len2;
-#ifdef HAVE_TSEARCH
-		    /*
-		     * If we are using tsearch(), make a local copy of the data
-		     * so we can trim it without interfering with tsearch's
-		     * notion of the ordering of data.  That will create some
-		     * spurious empty data, so we add the changed() macro in a
-		     * few places to skip over those.
-		     */
-		    if (use_tsearch) {
-			char *trim = new_string(data->name);
-			trim[len1] = EOS;
-			data = add_tsearch_data(trim, data->base);
-			free(trim);
-			local = 1;
-		    } else
-#endif
-			data->name[len1] = EOS;
+		if (!strncmp(target, source, len2)) {
+		    TRACE(("** trimming data \"%s\" to \"%.*s\"\n",
+			   target, (int) len2, target));
+		    if (reverse_opt) {
+			TRACE((".. no action @%d\n", __LINE__));
+		    } else {
+			target = trim_datapath(&data, len1 = len2, &local);
+		    }
 		}
 	    } else if (len1 < len2) {
-		if (!strncmp(data->name, path, len1)) {
-		    TRACE(("** trimming path '%s' to '%.*s'\n",
-			   path, (int) len1, path));
-		    path[len2 = len1] = EOS;
+		if (!strncmp(target, source, len1)) {
+		    TRACE(("** trimming source \"%s\" to \"%.*s\"\n",
+			   source, (int) len1, source));
+		    if (reverse_opt) {
+			TRACE((".. no action @%d\n", __LINE__));
+		    } else {
+			source[len2 = len1] = EOS;
+		    }
 		}
 	    }
 
-	    for (n = 1; n <= len1 && n <= len2; n++) {
-		if (data->name[len1 - n] != path[len2 - n]) {
-		    diff = (int) n;
-		    break;
+	    /*
+	     * If there was no "-p" option, look for the best match by
+	     * stripping prefixes from both source/target strings.
+	     */
+	    if (prefix_opt < 0) {
+		/*
+		 * Now (whether or not we trimmed a suffix), scan back from the
+		 * end of source/target strings to find if they happen to share
+		 * a common ending, e.g., a/b/c versus d/b/c.  If the strings
+		 * are not identical, then 'diff' will be set, but if they have
+		 * a common ending then 'matched' will be set.
+		 */
+		diff = 0;
+		matched = (int) compare_tails(target, source, &diff);
+
+		TRACE(("** merge @%d, prefix_opt=%d matched=%d diff=%d\n",
+		       __LINE__, prefix_opt, matched, diff));
+		if (matched != 0 && diff) {
+		    if (reverse_opt) {
+			TRACE((".. no action @%d\n", __LINE__));
+		    } else {
+			result = source + ((int) len2 - matched + 1);
+		    }
 		}
-		if (path[len2 - n] == PATHSEP)
-		    matched = (int) n;
 	    }
 
-	    if (prefix_opt < 0
-		&& matched != 0
-		&& diff)
-		path += ((int) len2 - matched + 1);
-
-	    if (!local)
-		*freed = delink(data);
-	    TRACE(("** merge @%d, prefix_opt=%d matched=%d diff=%d\n",
-		   __LINE__, prefix_opt, matched, diff));
-	} else if (!can_be_merged(path)) {
-	    TRACE(("** do not merge, retain @%d\n", __LINE__));
-	    /* must not merge, retain existing name */
-	    path = data->name;
+	    if (!local) {
+		if (reverse_opt) {
+		    TRACE((".. no action @%d\n", __LINE__));
+		} else {
+		    *freed = delink(data);
+		}
+	    }
+	} else if (reverse_opt) {
+	    TRACE((".. no action @%d\n", __LINE__));
+	    if (can_be_merged(source)) {
+		TRACE(("** merge @%d\n", __LINE__));
+	    } else {
+		TRACE(("** do not merge, retain @%d\n", __LINE__));
+		/* must not merge, retain existing name */
+		result = target;
+	    }
 	} else {
+	    if (can_be_merged(source)) {
+		TRACE(("** merge @%d\n", __LINE__));
+		*freed = delink(data);
+	    } else {
+		TRACE(("** do not merge, retain @%d\n", __LINE__));
+		/* must not merge, retain existing name */
+		result = target;
+	    }
+	}
+    } else if (reverse_opt) {
+	TRACE((".. no action @%d\n", __LINE__));
+	if (can_be_merged(source)) {
 	    TRACE(("** merge @%d\n", __LINE__));
-	    *freed = delink(data);
+	    result = target;
+	} else {
+	    TRACE(("** do not merge, retain @%d\n", __LINE__));
+	}
+    } else {
+	if (can_be_merged(source)) {
+	    TRACE(("** merge @%d\n", __LINE__));
+	    if (merge_names
+		&& *target != '\0'
+		&& prefix_opt < 0) {
+		size_t matched = compare_tails(target, source, &diff);
+		if (matched)
+		    result = target + (int) (strlen(target) - matched);
+	    }
+	} else {
+	    TRACE(("** do not merge, retain @%d\n", __LINE__));
+	    result = target;
 	}
-    } else if (!can_be_merged(path)) {
-	path = data->name;
     }
-    TRACE(("** finish do_merging ->%s\n", path));
-    return path;
+    TRACE(("** finish do_merging ->\"%s\"\n", result));
+    return result;
 }
 
 static int
 begin_data(const DATA * p)
 {
+    TRACE(("...begin_data(%s)\n", p->name));
     if (!can_be_merged(p->name)
 	&& strchr(p->name, PATHSEP) != 0) {
 	TRACE(("** begin_data:HAVE_PATH\n"));
@@ -822,10 +949,31 @@ skip_blanks(char *s)
     return s;
 }
 
+/*
+ * Skip a filename, which may be in quotes, to allow embedded blanks in the
+ * name.
+ */
+static char *
+skip_filename(char *s)
+{
+    if (*s == SQUOTE && s[1] != EOS && strchr(s + 1, SQUOTE)) {
+	++s;
+	while (*s != EOS && (*s != SQUOTE) && isgraph(UC(*s))) {
+	    ++s;
+	}
+	++s;
+    } else {
+	while (*s != EOS && isgraph(UC(*s))) {
+	    ++s;
+	}
+    }
+    return s;
+}
+
 static char *
 skip_options(char *params)
 {
-    while (*params != '\0') {
+    while (*params != EOS) {
 	params = skip_blanks(params);
 	if (*params == '-') {
 	    while (isgraph(UC(*params)))
@@ -834,7 +982,7 @@ skip_options(char *params)
 	    break;
 	}
     }
-    return params;
+    return skip_blanks(params);
 }
 
 /*
@@ -891,7 +1039,7 @@ get_line(char **buffer, size_t *have, FILE *fp)
 	if (ch == '\n')
 	    break;
     }
-    (*buffer)[used] = '\0';
+    (*buffer)[used] = EOS;
     return (used != 0);
 }
 
@@ -1042,6 +1190,7 @@ do_file(FILE *fp, const char *default_name)
     long new_dft = 0;
 
     int context = 1;
+    int either = 0;
 
     char *s;
 #if OPT_TRACE
@@ -1183,7 +1332,7 @@ do_file(FILE *fp, const char *default_name)
 		if (new_unify)
 		    --new_unify;
 		break;
-	    case '\0':
+	    case EOS:
 	    case ' ':
 		if (old_unify)
 		    --old_unify;
@@ -1281,8 +1430,7 @@ do_file(FILE *fp, const char *default_name)
 	     */
 	case 'I':
 	    CASE_TRACE();
-	    if (match(buffer, "Index: ")) {
-		s = strrchr(buffer, BLANK);	/* last token is name */
+	    if ((s = match(buffer, "Index: ")) != 0) {
 		s = skip_blanks(s);
 		dequote(s);
 		blip('.');
@@ -1295,10 +1443,14 @@ do_file(FILE *fp, const char *default_name)
 
 	case 'd':		/* diff command trace */
 	    CASE_TRACE();
-	    if (match(buffer, "diff ")
-		&& *skip_options(buffer + 5) != '\0') {
-		s = strrchr(buffer, BLANK);
-		s = skip_blanks(s);
+	    if ((s = match(buffer, "diff ")) != 0
+		&& *(s = skip_options(s)) != EOS) {
+		if (reverse_opt) {
+		    *skip_filename(s) = EOS;
+		} else {
+		    s = skip_filename(s);
+		    s = skip_blanks(s);
+		}
 		dequote(s);
 		blip('.');
 		finish_chunk(that);
@@ -1330,6 +1482,10 @@ do_file(FILE *fp, const char *default_name)
 			       &hour, &minute, &second) == 9
 			&& date_delims(yrmon, monday)
 			&& !version_num(b_fname))
+		    || (sscanf(buffer,
+			       "*** %[^\t]\t(%[^)])\t(%[^)])",
+			       b_fname, b_temp1, b_temp2) == 3
+			&& !version_num(b_fname))
 		    || sscanf(buffer,
 			      "*** %[^\t ]%[\t ]%[^ ] %[^ ] %d %d:%d:%d %d",
 			      b_fname,
@@ -1429,21 +1585,56 @@ do_file(FILE *fp, const char *default_name)
 	    }
 	    break;
 
-	    /* Expecting "Binary files XXX and YYY differ" */
-	case 'B':		/* Binary */
-	    /* FALL-THRU */
-	case 'b':		/* binary */
+	    /* Expecting "Files XXX and YYY differ" */
+	case 'F':		/* FALL-THRU */
+	case 'f':
 	    CASE_TRACE();
-	    if (match(buffer + 1, "inary files ")) {
+	    if ((s = match(buffer + 1, "iles ")) != 0) {
+		char *first = skip_blanks(s);
 		/* blindly assume the first filename does not contain " and " */
-		char *at_and = strstr(buffer, " and ");
+		char *at_and = strstr(s, " and ");
 		s = strrchr(buffer, BLANK);
 		if ((at_and != NULL) && !strcmp(s, " differ")) {
-		    *s = EOS;
-		    s = at_and + strlen(" and ");
+		    char *second = skip_blanks(at_and + 5);
+
+		    if (reverse_opt) {
+			*at_and = EOS;
+			s = first;
+		    } else {
+			*s = EOS;
+			s = second;
+		    }
 		    blip('.');
 		    finish_chunk(that);
-		    that = find_data(skip_blanks(s));
+		    that = find_data(s);
+		    that->cmt = Either;
+		    ok = HAVE_NOTHING;
+		    either = 1;
+		}
+	    }
+	    break;
+	    /* Expecting "Binary files XXX and YYY differ" */
+	case 'B':		/* FALL-THRU */
+	case 'b':
+	    CASE_TRACE();
+	    if ((s = match(buffer + 1, "inary files ")) != 0) {
+		char *first = skip_blanks(s);
+		/* blindly assume the first filename does not contain " and " */
+		char *at_and = strstr(s, " and ");
+		s = strrchr(buffer, BLANK);
+		if ((at_and != NULL) && !strcmp(s, " differ")) {
+		    char *second = skip_blanks(at_and + 5);
+
+		    if (reverse_opt) {
+			*at_and = EOS;
+			s = first;
+		    } else {
+			*s = EOS;
+			s = second;
+		    }
+		    blip('.');
+		    finish_chunk(that);
+		    that = find_data(s);
 		    that->cmt = Binary;
 		    ok = HAVE_NOTHING;
 		}
@@ -1455,13 +1646,39 @@ do_file(FILE *fp, const char *default_name)
 
     finish_chunk(that);
     finish_chunk(&dummy);
-    if (buffer != 0) {
-	free(buffer);
-	free(b_fname);
-	free(b_temp1);
-	free(b_temp2);
-	free(b_temp3);
+
+    if (either) {
+	int pass;
+	int fixup_diffs = 0;
+
+	for (pass = 0; pass < 2; ++pass) {
+	    DATA *p;
+	    for (p = all_data; p; p = p->link) {
+		switch (p->cmt) {
+		default:
+		    break;
+		case Normal:
+		    fixup_diffs = 1;
+		    break;
+		case Either:
+		    if (pass) {
+			if (fixup_diffs) {
+			    p->cmt = Binary;
+			} else {
+			    p->cmt = Differs;
+			}
+		    }
+		    break;
+		}
+	    }
+	}
     }
+
+    free(buffer);
+    free(b_fname);
+    free(b_temp1);
+    free(b_temp2);
+    free(b_temp3);
 }
 
 static void
@@ -1698,6 +1915,11 @@ show_data(const DATA * p)
 		   ModOf(p));
 	    if (path_opt)
 		printf("%ld,", EqlOf(p));
+	    if (count_files && !reverse_opt)
+		printf("%d,%d,%d,",
+		       (p->cmt == OnlyRight),
+		       (p->cmt == OnlyLeft),
+		       (p->cmt == Binary));
 	    printf("%s\n", name);
 	}
     } else if (names_only) {
@@ -1723,9 +1945,18 @@ show_data(const DATA * p)
 	case Binary:
 	    printf("binary");
 	    break;
+	case Differs:
+	    printf("differ");
+	    break;
 	case Only:
 	    printf("only");
 	    break;
+	case OnlyLeft:
+	    printf(count_files ? "deleted" : "only");
+	    break;
+	case OnlyRight:
+	    printf(count_files ? "added" : "only");
+	    break;
 	}
 	printf("\n");
     }
@@ -1749,6 +1980,51 @@ ignore_data(DATA * p)
 	    || (p->cmt == Binary && suppress_binary));
 }
 
+/*
+ * Return the length of any directory-prefix from the given path.
+ */
+static size_t
+path_length(const char *path)
+{
+    size_t result = 0;
+    char *mark = strrchr(path, PATHSEP);
+    if (mark != 0 && mark != path)
+	result = (size_t) (mark + 1 - path);
+    return result;
+}
+
+/*
+ * If we have an "only" filename, we can guess whether it was added or removed
+ * by looking at its directory and comparing that to other files' directories.
+ *
+ * TODO: -K -R combination is not yet supported because that relies on storing
+ * both left-/right-paths for each file; only the right-path is currently used.
+ */
+static Comment
+resolve_only(DATA * p)
+{
+    Comment result = p->cmt;
+    if (result == Only && !reverse_opt) {
+	DATA *q;
+	size_t len1 = path_length(p->name);
+	if (len1 != 0) {
+	    for (q = all_data; q; q = q->link) {
+		result = OnlyLeft;
+		if (q->cmt == Normal || q->cmt == Binary) {
+		    size_t len2 = path_length(q->name);
+		    if (len2 >= len1) {
+			if (!strncmp(p->name, q->name, len1)) {
+			    result = OnlyRight;
+			    break;
+			}
+		    }
+		}
+	    }
+	}
+    }
+    return result;
+}
+
 static void
 summarize(void)
 {
@@ -1757,6 +2033,9 @@ summarize(void)
     long total_del = 0;
     long total_mod = 0;
     long total_eql = 0;
+    long files_added = 0;
+    long files_binary = 0;
+    long files_removed = 0;
     long temp;
     int num_files = 0, shortest_name = -1, longest_name = -1;
 
@@ -1800,6 +2079,35 @@ summarize(void)
 	}
     }
 
+    /*
+     * Get additional counts for files where we cannot count lines changed.
+     */
+    if (count_files) {
+	for (p = all_data; p; p = p->link) {
+	    switch (p->cmt) {
+	    case Binary:
+		files_binary++;
+		break;
+	    case Only:
+		switch (resolve_only(p)) {
+		case OnlyRight:
+		    p->cmt = OnlyRight;
+		    files_added++;
+		    break;
+		case OnlyLeft:
+		    p->cmt = OnlyLeft;
+		    files_removed++;
+		    break;
+		default:
+		    /* ignore - we could not guess */
+		    break;
+		}
+	    default:
+		break;
+	    }
+	}
+    }
+
     /*
      * Use a separate loop after computing prefix_len so we can apply the "-S"
      * or "-D" options to find files that we can use as reference for the
@@ -1808,6 +2116,12 @@ summarize(void)
     for (p = all_data; p; p = p->link) {
 	if (!ignore_data(p)) {
 	    EqlOf(p) = 0;
+	    if (reverse_opt) {
+		long save_ins = InsOf(p);
+		long save_del = DelOf(p);
+		InsOf(p) = save_del;
+		DelOf(p) = save_ins;
+	    }
 	    if (path_opt != 0) {
 		int count = count_lines(p);
 
@@ -1853,6 +2167,8 @@ summarize(void)
 	    printf("INSERTED,DELETED,MODIFIED,");
 	    if (path_opt)
 		printf("UNCHANGED,");
+	    if (count_files && !reverse_opt)
+		printf("FILE-ADDED,FILE-DELETED,FILE-BINARY,");
 	}
 	printf("FILENAME\n");
     }
@@ -1877,6 +2193,14 @@ summarize(void)
 		printf(", %ld modification%s(!)", PLURAL(total_mod));
 	    if (total_eql && path_opt != 0)
 		printf(", %ld unchanged line%s(=)", PLURAL(total_eql));
+	    if (count_files) {
+		if (files_added)
+		    printf(", %ld file%s added", PLURAL(files_added));
+		if (files_removed)
+		    printf(", %ld file%s removed", PLURAL(files_removed));
+		if (files_binary)
+		    printf(", %ld binary file%s", PLURAL(files_binary));
+	    }
 	    (void) putchar('\n');
 	}
     }
@@ -1887,7 +2211,7 @@ static const char *
 get_program(const char *name, const char *dft)
 {
     const char *result = getenv(name);
-    if (result == 0 || *result == '\0')
+    if (result == 0 || *result == EOS)
 	result = dft;
     TRACE(("get_program(%s) = %s\n", name, result));
     return result;
@@ -1969,7 +2293,7 @@ is_compressed(const char *name)
     } else if (len > 5 && !strcmp(name + len - 5, ".lzma")) {
 	which = dcLzma;
     } else if (len > 3 && !strcmp(name + len - 3, ".xz")) {
-	which = dcLzma;
+	which = dcXz;
     } else {
 	which = dcNone;
     }
@@ -2062,7 +2386,9 @@ usage(FILE *fp)
 	"If no filename is given on the command line, reads from standard input.",
 	"",
 	"Options:",
+	"  -b      ignore lines matching \"Binary files XXX and YYY differ\"",
 	"  -c      prefix each line with comment (#)",
+	"  -C      add SGR color escape sequences to highlight the histogram",
 #if OPT_TRACE
 	"  -d      debug - prints a lot of information",
 #endif
@@ -2071,6 +2397,7 @@ usage(FILE *fp)
 	"  -f NUM  format (0=concise, 1=normal, 2=filled, 4=values)",
 	"  -h      print this message",
 	"  -k      do not merge filenames",
+	"  -K      resolve ambiguity of \"only\" filenames",
 	"  -l      list filenames only",
 	"  -m      merge insert/delete data in chunks as modified-lines",
 	"  -n NUM  specify minimum width for the filenames (default: auto)",
@@ -2079,6 +2406,8 @@ usage(FILE *fp)
 	"  -p NUM  specify number of pathname-separators to strip (default: common)",
 	"  -q      suppress the \"0 files changed\" message for empty diffs",
 	"  -r NUM  specify rounding for histogram (0=none, 1=simple, 2=adjusted)",
+	"  -R      assume patch was created with old and new files swapped",
+	"  -s      show only the summary line",
 	"  -S PATH specify location of original files, use for unchanged-count",
 	"  -t      print a table (comma-separated-values) rather than histogram",
 	"  -u      do not sort the input list",
@@ -2111,6 +2440,18 @@ getopt_helper(int argc, char *const argv[], const char *opts,
     return getopt(argc, argv, opts);
 }
 
+static int
+getopt_value(void)
+{
+    char *next = 0;
+    long value = strtol(optarg, &next, 0);
+    if (next == 0 || *next != '\0') {
+	fprintf(stderr, "expected a number, have '%s'\n", optarg);
+	exit(EXIT_FAILURE);
+    }
+    return (int) value;
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -2120,7 +2461,7 @@ main(int argc, char *argv[])
     max_width = 80;
 
     while ((j = getopt_helper(argc, argv,
-			      "bcCdD:e:f:hklmn:N:o:p:qr:sS:tuvVw:", 'h', 'V'))
+			      "bcCdD:e:f:hkKlmn:N:o:p:qr:RsS:tuvVw:", 'h', 'V'))
 	   != -1) {
 	switch (j) {
 	case 'b':
@@ -2145,7 +2486,7 @@ main(int argc, char *argv[])
 		failed(optarg);
 	    break;
 	case 'f':
-	    format_opt = atoi(optarg);
+	    format_opt = getopt_value();
 	    break;
 	case 'h':
 	    usage(stdout);
@@ -2153,6 +2494,9 @@ main(int argc, char *argv[])
 	case 'k':
 	    merge_names = 0;
 	    break;
+	case 'K':
+	    count_files = 1;
+	    break;
 	case 'l':
 	    names_only = 1;
 	    break;
@@ -2160,20 +2504,23 @@ main(int argc, char *argv[])
 	    merge_opt = 1;
 	    break;
 	case 'n':
-	    min_name_wide = atoi(optarg);
+	    min_name_wide = getopt_value();
 	    break;
 	case 'N':
-	    max_name_wide = atoi(optarg);
+	    max_name_wide = getopt_value();
 	    break;
 	case 'o':
 	    if (freopen(optarg, "w", stdout) == 0)
 		failed(optarg);
 	    break;
 	case 'p':
-	    prefix_opt = atoi(optarg);
+	    prefix_opt = getopt_value();
 	    break;
 	case 'r':
-	    round_opt = atoi(optarg);
+	    round_opt = getopt_value();
+	    break;
+	case 'R':
+	    reverse_opt = 1;
 	    break;
 	case 's':
 	    summary_only = 1;
@@ -2198,7 +2545,7 @@ main(int argc, char *argv[])
 	    printf("diffstat version %s\n", version);
 	    return (EXIT_SUCCESS);
 	case 'w':
-	    max_width = atoi(optarg);
+	    max_width = getopt_value();
 	    break;
 	case 'q':
 	    quiet = 1;
